/*
 * Copyright (c) 2023 Ryota Sasaki.
 *
 * This source code is licensed under the MIT license found in the LICENSE file in the root
 * directory of this source tree.
 */

#pragma once

#include <string>
#include <string_view>
#include <type_traits>

#include "../escape_sequence.hpp"
#include "../export_command/export_command.hpp"
#include "../options.hpp"
#include "../type_check.hpp"
#include "../utility.hpp"
#include "./export_var_fwd.hpp"

namespace cpp_dump {

namespace _detail {

template <typename T>
struct _map_dummy_wrapper {
 public:
  _map_dummy_wrapper(const T &map) : _map(map) {}

  auto begin() const noexcept { return _map.begin(); }
  auto end() const noexcept { return _map.end(); }
  auto size() const noexcept { return _map.size(); }

 private:
  const T &_map;
};

template <typename T>
struct _multimap_wrapper {
 public:
  _multimap_wrapper(const T &map) : _begin(map, map.begin()), _end(map, map.end()) {}

  auto begin() const noexcept { return _begin; }
  auto end() const noexcept { return _end; }

 private:
  struct multimap_wrapper_iterator {
   public:
    using It = typename T::const_iterator;
    multimap_wrapper_iterator(const T &map, It it) : _map(map), _it(it) {}

    const auto &operator*() const noexcept { return *_it; }
    auto operator->() const noexcept { return _it.operator->(); }
    bool operator!=(const multimap_wrapper_iterator &to) const { return _it != to._it; }
    multimap_wrapper_iterator &operator++() {
      _it = _map.equal_range(_it->first).second;
      return *this;
    }

   private:
    const T &_map;
    It _it;
  };

  multimap_wrapper_iterator _begin;
  multimap_wrapper_iterator _end;
};

template <typename It>
struct _multimap_value_wrapper {
 public:
  _multimap_value_wrapper(It begin, It end) : _begin(begin), _end(end) {}
  _multimap_value_wrapper() = delete;

  auto begin() const noexcept { return _begin; }
  auto end() const noexcept { return _end; }

 private:
  struct multimap_value_iterator {
   public:
    multimap_value_iterator(It it) : _it(it) {}
    multimap_value_iterator() = delete;

    const auto &operator*() const noexcept { return _it->second; }
    bool operator!=(const multimap_value_iterator &to) const noexcept { return _it != to._it; }
    multimap_value_iterator &operator++() {
      ++_it;
      return *this;
    }

   private:
    It _it;
  };

  multimap_value_iterator _begin;
  multimap_value_iterator _end;
};

template <typename T>
inline auto export_map(
    const T &map,
    const std::string &indent,
    std::size_t last_line_length,
    std::size_t current_depth,
    bool fail_on_newline,
    const export_command &command
) -> std::enable_if_t<is_map<T>, std::string> {
  if (map.empty()) return es::bracket("{ }", current_depth);

  if (current_depth >= max_depth)
    return es::bracket("{ ", current_depth) + es::op("...") + es::bracket(" }", current_depth);

  bool shift_indent = is_multimap<T> || is_iterable_like<typename T::key_type>
                      || is_iterable_like<typename T::mapped_type>;

  if (shift_indent && fail_on_newline) return "\n";

  std::string new_indent = indent + "  ";
  std::size_t next_depth = current_depth + 1;
  const auto &key_command = command.next_for_map_key();
  const auto &value_command = command.next_for_map_value();

  auto map_wrapper = ([&]() {
    if constexpr (is_multimap<T>) {
      return _multimap_wrapper(map);
    } else {
      // The wrapper is to avoid calling the copy constructor.
      return _map_dummy_wrapper(map);
    }
  })();
  auto skipped_map = command.create_skip_container(map_wrapper);

rollback:
  std::string output = es::bracket("{ ", current_depth);
  bool is_first = true;

  for (const auto &[skip, it, _] : skipped_map) {
    const auto &[key, value] = *it;

    if (is_first) {
      is_first = false;
    } else {
      output += es::op(", ");
    }

    std::string key_string, value_string;
    if (shift_indent) {
      if (skip) {
        output += "\n" + new_indent + es::op("...");
        continue;
      }

      if constexpr (is_multimap<T>) {
        auto [_begin, _end] = map.equal_range(key);
        _multimap_value_wrapper values(_begin, _end);

        // Treat the multiplicity as a member to distinguish it from the keys & values.
        // Also, multiplicities are similar to members since they are on the left side of values.
        key_string =
            "\n" + new_indent
            + export_var(key, new_indent, new_indent.length(), next_depth, false, key_command)
            + es::member(" (" + std::to_string(map.count(key)) + ")") + es::op(": ");
        value_string = export_var(
            values, new_indent, get_last_line_length(key_string), next_depth, false, value_command
        );
      } else {
        key_string =
            "\n" + new_indent
            + export_var(key, new_indent, new_indent.length(), next_depth, false, key_command)
            + es::op(": ");
        value_string = export_var(
            value, new_indent, get_last_line_length(key_string), next_depth, false, value_command
        );
      }

      output += key_string + value_string;
      continue;
    }

    if (skip) {
      output += es::op("...");

      if (last_line_length + get_length(output) + std::string_view(" }").size() <= max_line_width)
        continue;

      shift_indent = true;
      goto rollback;
    }

    if constexpr (is_multimap<T>) {
      auto [_begin, _end] = map.equal_range(key);
      _multimap_value_wrapper values(_begin, _end);

      // Treat the multiplicity as a member to distinguish it from the keys & values.
      // Also, multiplicities are similar to members since they are on the left side of values.
      key_string =
          export_var(
              key, indent, last_line_length + get_length(output), next_depth, true, key_command
          )
          + es::member(" (" + std::to_string(map.count(key)) + ")") + es::op(": ");
      value_string = export_var(
          values,
          indent,
          last_line_length + get_length(output) + get_length(key_string),
          next_depth,
          true,
          value_command
      );
    } else {
      key_string =
          export_var(
              key, indent, last_line_length + get_length(output), next_depth, true, key_command
          )
          + es::op(": ");
      value_string = export_var(
          value,
          indent,
          last_line_length + get_length(output) + get_length(key_string),
          next_depth,
          true,
          value_command
      );
    }

    std::string elem_string = key_string + value_string;
    if (!has_newline(elem_string)) {
      output += elem_string;

      if (last_line_length + get_length(output) + std::string_view(" }").size() <= max_line_width)
        continue;
    }

    if (fail_on_newline) return "\n";

    shift_indent = true;
    goto rollback;
  }

  if (shift_indent) {
    output += "\n" + indent + es::bracket("}", current_depth);
  } else {
    output += es::bracket(" }", current_depth);
  }

  return output;
}

}  // namespace _detail

}  // namespace cpp_dump
